#include "geometry_msgs/msg/pose.hpp"
#include <filesystem>
#include <memory>
#include <rclcpp/logger.hpp>
#include <rclcpp/node.hpp>
#include <rclcpp/node_interfaces/node_base_interface.hpp>
#include <rclcpp/node_interfaces/node_logging_interface.hpp>
#include <string>
#include <tf2/LinearMath/Quaternion.hpp>
#include <vector>
#include <yaml-cpp/node/node.h>
#include <yaml-cpp/node/parse.h>
#include <yaml-cpp/yaml.h>

class PlacesConfig {
public:
  PlacesConfig(const std::string &places_filepath, const rclcpp::Logger& logger)
      : _logger(logger) {

    objects_to_spawn_ = std::make_shared<Places>();
    if (!loadPlacesFromYaml(places_filepath)) {
      RCLCPP_FATAL(_logger, "Failed to load objects from YAML file: %s",
                   places_filepath.c_str());
      rclcpp::shutdown();
      return;
    }
  }

  PlacesConfig(const std::string &places_filepath, rclcpp::Node::SharedPtr node)
      : PlacesConfig(places_filepath, node->get_logger()) {}

  struct PlaceSpec {
    std::string name;
    std::string type;
    std::string frame_id = "world";
    bool has_dimensions;
    std::vector<double> dimensions;
    geometry_msgs::msg::Pose pose;
    std::optional<std::string> mesh_filepath;
    std::tuple<double, double, double> scale_xyz = {1.0, 1.0, 1.0};
  };

  struct Places {
    std::optional<std::string> model_dir;
    std::vector<PlaceSpec> models;
  };

  inline std::shared_ptr<Places> get() const { return objects_to_spawn_; }

private:
  const rclcpp::Logger& _logger;

  bool loadPlacesFromYaml(const std::string &filepath) {
    auto places_config_node = YAML::LoadFile(filepath);

    if (auto node = places_config_node["model_dir"]; node && node.IsScalar()) {
      objects_to_spawn_->model_dir = node.as<std::string>();
    }

    if (!places_config_node["models"] ||
        !places_config_node["models"].IsSequence()) {
      RCLCPP_FATAL(_logger, "No valid 'models' section found in config_file");
      return false;
    }
    for (const auto &model : places_config_node["models"]) {
      PlaceSpec object;
      if (model["ps_ignore"] && model["ps_ignore"].as<bool>()) {
        continue;
      }
      object.name = model["name"].as<std::string>();
      object.type = model["type"].as<std::string>();
      const auto &type = object.type;
      if (type == "box" || type == "sphere" || type == "cylinder") {
        if (!model["dimensions"]) {
          RCLCPP_ERROR(_logger, "Model '%s' is missing 'dimensions'",
                       object.name.c_str());
          return false;
        }
        auto d = model["dimensions"].as<std::vector<double>>();

        if (type == "box" && d.size() == 3) {
          object.dimensions = d;
        } else if (type == "sphere" && d.size() == 1) {
          object.dimensions = d;
        } else if (type == "cylinder" && d.size() == 2) {
          object.dimensions = d;
        } else {
          RCLCPP_ERROR(_logger, "Invalid dimensions count for type '%s'",
                       type.c_str());
          return false;
        }

      } else if (type == "mesh") {
        if (!model["mesh_filename"]) {
          RCLCPP_ERROR(_logger, "mesh object missing mesh_filename");
          return false;
        }
        std::string mesh_filename = model["mesh_filename"].as<std::string>();
        if (objects_to_spawn_->model_dir.has_value()) {
          object.mesh_filepath =
              std::filesystem::path(objects_to_spawn_->model_dir->c_str()) /
              object.name / "meshes" / mesh_filename;
          RCLCPP_WARN(_logger, "mesh filepath is: %s",
                      object.mesh_filepath->c_str());
        } else {
          RCLCPP_ERROR(_logger, "mesh_filename in empty in config_file");
          return false;
        }

      } else {
        RCLCPP_ERROR(_logger, "Unrecognized object type: '%s'", type.c_str());
        return false;
      }
      if (model["frame_id"]) {
        object.frame_id = model["frame_id"].as<std::string>();
      }
      if (model["pose"]) {
        auto pose = model["pose"];

        if (pose["position"]) {
          auto position = model["pose"]["position"];
          object.pose.position.x = position["x"].as<double>();
          object.pose.position.y = position["y"].as<double>();
          object.pose.position.z = position["z"].as<double>();
        }
        if (pose["orientation"]) {
          auto orientation = model["pose"]["orientation"];
          if (!isQuat(orientation)) {
            auto quat = rpyToQuat(orientation["r"].as<double>(),
                                  orientation["p"].as<double>(),
                                  orientation["y"].as<double>());

            object.pose.orientation.x = quat.x;
            object.pose.orientation.y = quat.y;
            object.pose.orientation.z = quat.z;
            object.pose.orientation.w = quat.w;
          } else {
            object.pose.orientation.x = orientation["x"].as<double>();
            object.pose.orientation.y = orientation["y"].as<double>();
            object.pose.orientation.z = orientation["z"].as<double>();
            object.pose.orientation.w = orientation["w"].as<double>();
          }
        }
      }

      objects_to_spawn_->models.push_back(object);
    }

    return true;
  }

  bool isQuat(const YAML::Node &node) {
    if (node["x"] && node["y"] && node["z"] && node["w"]) {
      return true;
    } else if (node["r"] && node["p"] && node["y"]) {
      return false;
    }
    throw std::invalid_argument(
        "Provided YAML Node is not quat or euler angles");
  }

  geometry_msgs::msg::Quaternion rpyToQuat(double roll, double pitch,
                                           double yaw) {
    tf2::Quaternion q;
    q.setRPY(roll, pitch, yaw);
    q.normalize();
    geometry_msgs::msg::Quaternion q_msg;
    q_msg.x = q.x();
    q_msg.y = q.y();
    q_msg.z = q.z();
    q_msg.w = q.w();
    return q_msg;
  }

  std::shared_ptr<Places> objects_to_spawn_;
};
