#include "geometry_msgs/msg/pose.hpp"
#include <filesystem>
#include <rclcpp/logger.hpp>
#include <rclcpp/node.hpp>
#include <stdexcept>
#include <string>
#include <tf2/LinearMath/Quaternion.hpp>
#include <yaml-cpp/yaml.h>

class RefFramesConfig {
public:
  RefFramesConfig(const std::string &ref_frames_filepath) {
    loadConfig(ref_frames_filepath);
  }

  struct RefFrame {
    std::string name;
    std::string relative_to;
    geometry_msgs::msg::Pose pose;
  };

  inline std::vector<RefFrame> get() const {
    return _ref_frames;
  }


private:
  std::vector<RefFrame> _ref_frames;

  geometry_msgs::msg::Quaternion toQuat(const YAML::Node &node) {
    geometry_msgs::msg::Quaternion quat;
    if (node["x"] && node["y"] && node["z"] && node["w"]) {
      quat.x = node["x"].as<double>();
      quat.y = node["y"].as<double>();
      quat.z = node["z"].as<double>();
      quat.w = node["w"].as<double>();
    } else if (node["r"] && node["p"] && node["y"]) {
      tf2::Quaternion tf2_quat;
      tf2_quat.setRPY(node["r"].as<double>(), node["p"].as<double>(),
                      node["y"].as<double>());
      quat.x = tf2_quat.x();
      quat.y = tf2_quat.y();
      quat.z = tf2_quat.z();
      quat.w = tf2_quat.w();
    }
    return quat;
  }

  void parseRefFrames(const YAML::Node &node) {
    for (const auto &frame : node) {
      RefFrame ref_frame;
      if (!frame["name"] || !frame["relative_to"] || !frame["pose"]) {
        throw std::runtime_error(
            "Missing required ref_frame parameters in YAML");
      }
      ref_frame.name = frame["name"].as<std::string>();
      ref_frame.relative_to = frame["relative_to"].as<std::string>();

      geometry_msgs::msg::Pose frame_pose;
      frame_pose.position.x = frame["pose"]["position"]["x"].as<double>();
      frame_pose.position.y = frame["pose"]["position"]["y"].as<double>();
      frame_pose.position.z = frame["pose"]["position"]["z"].as<double>();

      if (auto orientation = frame["pose"]["orientation"]; orientation) {
        frame_pose.orientation = toQuat(orientation);
      }
      ref_frame.pose = frame_pose;
      _ref_frames.push_back(ref_frame);
    }
  }

  void loadConfig(const std::string &ref_frames_filepath) {
    YAML::Node config = YAML::LoadFile(ref_frames_filepath);

    if (config["ref_frames"] && config["ref_frames"].IsSequence()) {
      parseRefFrames(config["ref_frames"]);
    } else {
      throw std::runtime_error("No ref_frames in config");
    }

  }
};
