#include "gz/sim/components/Name.hh"
#include "std_msgs/msg/bool.hpp"
#include <functional>
#include <gz/common/Console.hh>
#include <gz/plugin/Register.hh>
#include <gz/sim/Entity.hh>
#include <gz/sim/EntityComponentManager.hh>
#include <gz/sim/Model.hh>

#include <gz/sim/components/ChildLinkName.hh>
#include <gz/sim/components/Collision.hh>
#include <gz/sim/components/ContactSensorData.hh>
#include <gz/sim/components/DetachableJoint.hh>
#include <gz/sim/components/Joint.hh>
#include <gz/sim/components/ParentEntity.hh>

#include <memory>
#include <rclcpp/rclcpp.hpp>
#include <string>

#include "rbs_mill_assist/VacuumGripper.hpp"

GZ_ADD_PLUGIN(rbs_mill_assist::VacuumGripper, gz::sim::System,
              rbs_mill_assist::VacuumGripper::ISystemConfigure,
              rbs_mill_assist::VacuumGripper::ISystemPreUpdate,
              rbs_mill_assist::VacuumGripper::ISystemPostUpdate)

namespace rbs_mill_assist {

class VacuumGripperPrivate {
public:
  std::string gripper_link_name;
  bool attach_request;
  bool detach_request;
  bool enabled;
  bool attached;
  gz::sim::EntityComponentManager _ecm;

  gz::sim::components::Joint joint;
  gz::sim::Entity contactedEntity = gz::sim::kNullEntity;
  gz::sim::Entity gripperEntity = gz::sim::kNullEntity;
  std::unordered_map<std::string, gz::sim::Entity> detachableJoints;

  std::shared_ptr<rclcpp::Node> node_{nullptr};
  std::thread thread_executor_spin_;
  rclcpp::executors::MultiThreadedExecutor::SharedPtr executor_;
};

VacuumGripper::VacuumGripper()
    : dataPtr(std::make_unique<VacuumGripperPrivate>()) {}

VacuumGripper::~VacuumGripper() {
  this->dataPtr->executor_->cancel();
  this->dataPtr->thread_executor_spin_.join();
}

void VacuumGripper::Toggle(
    const std::shared_ptr<std_srvs::srv::SetBool::Request> request,
    std::shared_ptr<std_srvs::srv::SetBool::Response> response) {
  this->dataPtr->enabled = request->data;
  response->success = true;
  gzdbg << "Vacuum Gripper " << (request->data ? "Activated" : "Deactivated")
        << std::endl;
}

bool VacuumGripper::AttachLink(gz::sim::EntityComponentManager &_ecm,
                               const gz::sim::Entity &gripperEntity,
                               const gz::sim::Entity &objectEntity) {

  if (gripperEntity == gz::sim::kNullEntity ||
      objectEntity == gz::sim::kNullEntity) {
    gzerr << "Invalid gripper or object entity!" << std::endl;
    return false;
  }

  auto jointEntity = _ecm.CreateEntity();

  gz::sim::components::DetachableJointInfo jointInfo;
  jointInfo.parentLink = gripperEntity;
  jointInfo.childLink = objectEntity;
  jointInfo.jointType = "fixed";

  _ecm.CreateComponent(jointEntity,
                       gz::sim::components::DetachableJoint(jointInfo));

  std::string key =
      std::to_string(gripperEntity) + "_" + std::to_string(objectEntity);
  this->dataPtr->detachableJoints[key] = jointEntity;

  gzdbg << "Attached object " << objectEntity << " to gripper " << gripperEntity
        << std::endl;

  this->dataPtr->contactedEntity = objectEntity;
  this->dataPtr->attached = true;

  return true;
}

bool VacuumGripper::DetachLink(gz::sim::EntityComponentManager &_ecm) {
  if (this->dataPtr->contactedEntity == gz::sim::kNullEntity) {
    gzdbg << "No object is currently attached to the gripper." << std::endl;
    return false;
  }

  std::string key = std::to_string(this->dataPtr->gripperEntity) + "_" +
                    std::to_string(this->dataPtr->contactedEntity);

  auto it = this->dataPtr->detachableJoints.find(key);
  if (it == this->dataPtr->detachableJoints.end()) {
    gzerr << "Failed to find detachable joint for object "
          << this->dataPtr->contactedEntity << std::endl;
    return false;
  }

  auto jointEntity = it->second;
  _ecm.RequestRemoveEntity(jointEntity);

  this->dataPtr->contactedEntity = gz::sim::kNullEntity;
  this->dataPtr->detachableJoints.erase(it);
  this->dataPtr->attached = false;

  return true;
}

void VacuumGripper::Configure(
    const gz::sim::Entity &_entity,
    const std::shared_ptr<const sdf::Element> &_element,
    gz::sim::EntityComponentManager &_ecm,
    gz::sim::EventManager &_eventManager) {

  rclcpp::Logger logger = rclcpp::get_logger("VacuumGripperGazeboPlugin");
  const auto model = gz::sim::Model(_entity);

  if (!model.Valid(_ecm)) {
    RCLCPP_ERROR(logger,
                 "[Vacuum Gripper Plugin] Failed to initialize because [%s] "
                 "(Entity=%lu)] is not a model."
                 "Please make sure that Vacuum Gripper Plugin is attached to a "
                 "valid model.",
                 model.Name(_ecm).c_str(), _entity);
    return;
  }

  auto sdfPtr = const_cast<sdf::Element *>(_element.get());

  if (!sdfPtr->HasElement("link_name")) {
    RCLCPP_ERROR(logger, "VacuumGripper link has not defined");
    return;
  }

  this->dataPtr->gripper_link_name =
      sdfPtr->GetElement("link_name")->Get<std::string>();

  _ecm.Each<gz::sim::components::Name>(
      [&](const gz::sim::Entity &entity,
          const gz::sim::components::Name *nameComp) -> bool {
        if (nameComp->Data() == this->dataPtr->gripper_link_name) {
          this->dataPtr->gripperEntity = entity;
          gzdbg << "Found gripper link: " << this->dataPtr->gripper_link_name
                << " (Entity ID: " << entity << ")" << std::endl;
          return false;
        }
        return true;
      });

  if (this->dataPtr->gripperEntity == gz::sim::kNullEntity) {
    RCLCPP_ERROR(logger, "Gripper link '%s' not found!",
                 this->dataPtr->gripper_link_name.c_str());
    return;
  }

  std::string node_name = "gz_ros2_vacuum_gripper_plugin";

  this->dataPtr->node_ = rclcpp::Node::make_shared(node_name);

  this->dataPtr->executor_ =
      std::make_shared<rclcpp::executors::MultiThreadedExecutor>();
  this->dataPtr->executor_->add_node(this->dataPtr->node_);
  auto spin = [this]() { this->dataPtr->executor_->spin(); };
  this->dataPtr->thread_executor_spin_ = std::thread(spin);

  toggle_service_ =
      this->dataPtr->node_->create_service<std_srvs::srv::SetBool>(
          node_name + "/toggle",
          std::bind(&VacuumGripper::Toggle, this, std::placeholders::_1,
                    std::placeholders::_2));

  contact_publisher_ =
      this->dataPtr->node_->create_publisher<std_msgs::msg::Bool>(
          node_name + "/contact_info", 10);

  gzdbg << "rbs_mill_assist::VacuumGripper::Configure on entity: " << _entity
        << std::endl;
}

void VacuumGripper::PreUpdate(const gz::sim::UpdateInfo &_info,
                              gz::sim::EntityComponentManager &_ecm) {
  if (!this->dataPtr->enabled) {
    if (this->dataPtr->contactedEntity != gz::sim::kNullEntity) {
      gzdbg << "Detaching object..." << std::endl;
      if (this->DetachLink(_ecm)) {
        gzdbg << "Object detached successfully!" << std::endl;
      } else {
        gzerr << "Failed to detach object!" << std::endl;
      }
    }
    return;
  }

  _ecm.Each<gz::sim::components::ContactSensorData>(
      [&](const gz::sim::Entity &entity,
          const gz::sim::components::ContactSensorData *contactData) -> bool {
        if (contactData->Data().contact_size() == 0) {
          return true;
        }

        auto msgEntity = contactData->Data().contact(0).collision2();
        gz::sim::Entity contactedCollision(msgEntity.id());

        auto parentLink = _ecm.Component<gz::sim::components::ParentEntity>(
            contactedCollision);
        if (!parentLink) {
          return true;
        }

        if (this->dataPtr->contactedEntity != gz::sim::kNullEntity) {
          return true;
        }

        gz::sim::Entity gripperLink = this->dataPtr->gripperEntity;
        gz::sim::Entity objectLink = parentLink->Data();

        gzdbg << "Attaching object " << objectLink << " to gripper..."
              << std::endl;
        if (this->AttachLink(_ecm, gripperLink, objectLink)) {
          this->dataPtr->contactedEntity = objectLink;
          gzdbg << "Object attached successfully!" << std::endl;
        } else {
          gzerr << "Failed to attach object!" << std::endl;
        }

        return true;
      });
}

void VacuumGripper::PostUpdate(const gz::sim::UpdateInfo &_info,
                               const gz::sim::EntityComponentManager &_ecm) {
  std_msgs::msg::Bool msg;
  msg.data = this->dataPtr->attached;
  contact_publisher_->publish(msg);
}

} // namespace rbs_mill_assist
