#include <chrono>
#include <ctime>
#include <memory>
#include <rclcpp/executors/single_threaded_executor.hpp>

#include "geometry_msgs/msg/pose.hpp"
#include "geometry_msgs/msg/transform_stamped.hpp"
#include "rbs_mill_assist/places_config.hpp"
#include "rbs_mill_assist/ref_frames_config.hpp"
#include "rclcpp/rclcpp.hpp"
#include "tf2/LinearMath/Quaternion.h"
#include "tf2_ros/static_transform_broadcaster.h"

class PlacesFramePublisher : public rclcpp::Node {
public:
  explicit PlacesFramePublisher() : Node("static_turtle_tf2_broadcaster") {
    tf_static_broadcaster_ =
        std::make_shared<tf2_ros::StaticTransformBroadcaster>(this);
    this->declare_parameter("places_config_filepath", "");
    this->declare_parameter("ref_frames_config_filepath", "");
    auto places_config_filepath =
        this->get_parameter("places_config_filepath").as_string();
    auto ref_frames_config_filepath =
        this->get_parameter("ref_frames_config_filepath").as_string();
    auto places_config = std::make_shared<PlacesConfig>(places_config_filepath, this->get_logger());
    auto ref_frames =
        std::make_shared<RefFramesConfig>(ref_frames_config_filepath);

    this->publish_places_frames(places_config);
    this->publish_ref_frames(ref_frames);
  }

private:
  void publish_ref_frames(std::shared_ptr<RefFramesConfig> &ref_frames_config) {
    auto ref_frames = ref_frames_config->get();
    for (size_t i = 0; i < ref_frames.size(); i++) {
      geometry_msgs::msg::TransformStamped t;

      t.header.stamp = this->get_clock()->now();
      t.header.frame_id = ref_frames[i].relative_to;
      t.child_frame_id = ref_frames[i].name;

      geometry_msgs::msg::Pose tmp_pose;
      tmp_pose = ref_frames[i].pose;

      t.transform.translation.x = tmp_pose.position.x;
      t.transform.translation.y = tmp_pose.position.y;
      t.transform.translation.z = tmp_pose.position.z;
      t.transform.rotation.x = tmp_pose.orientation.x;
      t.transform.rotation.y = tmp_pose.orientation.y;
      t.transform.rotation.z = tmp_pose.orientation.z;
      t.transform.rotation.w = tmp_pose.orientation.w;

      tf_static_broadcaster_->sendTransform(t);
    }
  }

  void publish_places_frames(std::shared_ptr<PlacesConfig> places_config) {
    auto places = places_config->get();
    for (size_t i = 0; i < places->models.size(); i++) {
      geometry_msgs::msg::TransformStamped t;

      t.header.stamp = this->get_clock()->now();
      t.header.frame_id = places->models[i].frame_id;
      t.child_frame_id = places->models[i].name;

      geometry_msgs::msg::Pose tmp_pose;
      tmp_pose = places->models[i].pose;

      t.transform.translation.x = tmp_pose.position.x;
      t.transform.translation.y = tmp_pose.position.y;
      t.transform.translation.z = tmp_pose.position.z;
      t.transform.rotation.x = tmp_pose.orientation.x;
      t.transform.rotation.y = tmp_pose.orientation.y;
      t.transform.rotation.z = tmp_pose.orientation.z;
      t.transform.rotation.w = tmp_pose.orientation.w;

      tf_static_broadcaster_->sendTransform(t);
    }
  }

  std::shared_ptr<tf2_ros::StaticTransformBroadcaster> tf_static_broadcaster_;
};

int main(int argc, char *argv[]) {
  rclcpp::init(argc, argv);
  rclcpp::executors::SingleThreadedExecutor executor;
  auto node = std::make_shared<PlacesFramePublisher>();
  executor.add_node(node);
  executor.spin();
  rclcpp::shutdown();
  return 0;
}
