#include "geometry_msgs/msg/pose.hpp"
#include "geometry_msgs/msg/quaternion.hpp"
#include "moveit_msgs/msg/collision_object.hpp"
#include "moveit_msgs/msg/planning_scene.hpp"
#include "shape_msgs/msg/mesh.hpp"
#include <chrono>
#include <functional>
#include <geometry_msgs/msg/pose.hpp>
#include <memory>
#include <rclcpp/duration.hpp>
#include <rclcpp/executors/single_threaded_executor.hpp>
#include <rclcpp/logging.hpp>
#include <rclcpp/publisher.hpp>
#include <rclcpp/rclcpp.hpp>

#include "rbs_mill_assist/places_config.hpp"
#include <geometric_shapes/shape_operations.h>
#include <moveit_msgs/msg/collision_object.hpp>
#include <moveit_msgs/msg/planning_scene.hpp>
#include <shape_msgs/msg/mesh.hpp>
#include <shape_msgs/msg/solid_primitive.hpp>
#include <tf2/time.hpp>
#include <tf2_ros/buffer.h>
#include <tf2_ros/transform_listener.h>

class PlanningSceneNode : public rclcpp::Node {
public:
  PlanningSceneNode()
      : Node("planning_scene_publisher"), object_published_(false) {
    this->declare_parameter("places_config", "");
    _planning_scene_publisher =
        this->create_publisher<moveit_msgs::msg::PlanningScene>(
            "planning_scene", 10);
    _timer = this->create_wall_timer(
        std::chrono::seconds(1),
        std::bind(&PlanningSceneNode::publish_scene, this));
    auto places_config_filepath =
        this->get_parameter("places_config").as_string();

    tf_buffer_ = std::make_unique<tf2_ros::Buffer>(this->get_clock());
    tf_listener_ = std::make_shared<tf2_ros::TransformListener>(*tf_buffer_);



    try {
      _places_config = std::make_shared<PlacesConfig>(places_config_filepath);
    } catch (const std::exception &e) {
      RCLCPP_ERROR(this->get_logger(), "Failed to load config: %s", e.what());
    }
  }

private:
  rclcpp::Publisher<moveit_msgs::msg::PlanningScene>::SharedPtr
      _planning_scene_publisher;
  rclcpp::TimerBase::SharedPtr _timer;
  bool object_published_;
  std::shared_ptr<PlacesConfig> _places_config;

  std::shared_ptr<tf2_ros::TransformListener> tf_listener_;
  std::shared_ptr<tf2_ros::Buffer> tf_buffer_;

  void publish_scene() {
    if (object_published_) {
      return;
    }

    moveit_msgs::msg::PlanningScene planning_scene;
    moveit_msgs::msg::CollisionObject collision_object;

    collision_object.header.frame_id = "world";
    collision_object.id = "table";

    shape_msgs::msg::SolidPrimitive table;
    table.type = shape_msgs::msg::SolidPrimitive::BOX;
    table.dimensions = {1.2, 0.7, 0.8};

    geometry_msgs::msg::Pose table_pose;
    table_pose.position.x = 0.0;
    table_pose.position.y = 0.0;
    table_pose.position.z = -0.4;

    collision_object.primitives.push_back(table);
    collision_object.primitive_poses.push_back(table_pose);
    collision_object.operation = moveit_msgs::msg::CollisionObject::ADD;

    planning_scene.world.collision_objects.push_back(collision_object);

    for (size_t i = 0; i < _places_config->getMeshFilepaths().size(); i++) {
      if (!tf_buffer_->canTransform(_places_config->getModelParents()[i],
                                    _places_config->getModelNames()[i],
                                    tf2::TimePointZero)) {
        RCLCPP_WARN(this->get_logger(), "Waiting for TF frame '%s'...",
                    _places_config->getModelParents()[i].c_str());
        return;
      }
      moveit_msgs::msg::CollisionObject mesh_object;
      mesh_object.id = _places_config->getModelNames()[i];
      mesh_object.header.frame_id = _places_config->getModelParents()[i];
      auto mesh = getMeshCollisionObject(_places_config->getMeshFilepaths()[i]);
      mesh_object.meshes.push_back(mesh);
      mesh_object.mesh_poses.push_back(_places_config->getMeshPoses()[i]);
      mesh_object.operation = moveit_msgs::msg::CollisionObject::ADD;
      planning_scene.world.collision_objects.push_back(mesh_object);
    }

    planning_scene.is_diff = true;

    _planning_scene_publisher->publish(planning_scene);
    RCLCPP_INFO(this->get_logger(),
                "Published collision objects: table and mesh");

    object_published_ = true;
  }

  shape_msgs::msg::Mesh
  getMeshCollisionObject(const std::string &mesh_filepath) {
    shape_msgs::msg::Mesh mesh;
    shapes::Mesh *m = shapes::createMeshFromResource(mesh_filepath);
    shapes::ShapeMsg mesh_msg;
    shapes::constructMsgFromShape(m, mesh_msg);
    mesh = boost::get<shape_msgs::msg::Mesh>(mesh_msg);
    return mesh;
  }
};

int main(int argc, char **argv) {
  rclcpp::init(argc, argv);
  rclcpp::executors::SingleThreadedExecutor executor;
  auto node = std::make_shared<PlanningSceneNode>();
  executor.add_node(node);
  executor.spin();
  rclcpp::shutdown();
  return 0;
}
