
#include "moveit_msgs/msg/collision_object.hpp"
#include "rbs_mill_interfaces/action/add_collision_object.hpp"
#include "rbs_mill_interfaces/action/attach_detach_object.hpp"
#include "rbs_mill_interfaces/action/check_object_exists.hpp"
#include "rbs_mill_interfaces/action/get_object_pose.hpp"
#include "rbs_mill_interfaces/action/remove_collision_object.hpp"
#include <geometry_msgs/msg/pose.hpp>
#include <geometry_msgs/msg/quaternion.hpp>
#include <moveit_msgs/msg/attached_collision_object.hpp>
#include <moveit_msgs/msg/collision_object.hpp>
#include <moveit_msgs/srv/get_planning_scene.hpp>
#include <rclcpp/callback_group.hpp>
#include <rclcpp/node_options.hpp>
#include <rclcpp/publisher.hpp>
#include <rclcpp/rclcpp.hpp>
#include <rclcpp/service.hpp>
#include <rclcpp_action/rclcpp_action.hpp>
#include <rclcpp_action/server.hpp>
#include <shape_msgs/msg/solid_primitive.hpp>

namespace rbs_mill_utils {

class PsManager : public rclcpp::Node {

public:
  using AttachDetachObject = rbs_mill_interfaces::action::AttachDetachObject;

  using AddCollisionObject = rbs_mill_interfaces::action::AddCollisionObject;
  using RemoveCollisionObject =
      rbs_mill_interfaces::action::RemoveCollisionObject;
  using CheckObjectExists = rbs_mill_interfaces::action::CheckObjectExists;
  using GetObjectPose = rbs_mill_interfaces::action::GetObjectPose;

  using AddGoalHandle = rclcpp_action::ServerGoalHandle<AddCollisionObject>;
  using RemoveGoalHandle =
      rclcpp_action::ServerGoalHandle<RemoveCollisionObject>;
  using CheckGoalHandle = rclcpp_action::ServerGoalHandle<CheckObjectExists>;
  using AttachDetachGoalHandle =
      rclcpp_action::ServerGoalHandle<AttachDetachObject>;
  using GetObjectPoseGoalHandle =
      rclcpp_action::ServerGoalHandle<GetObjectPose>;

  PsManager();

private:
  rclcpp_action::GoalResponse
  handleAddGoal(const rclcpp_action::GoalUUID &uuid,
                std::shared_ptr<const AddCollisionObject::Goal> goal);

  rclcpp_action::CancelResponse
  handleAddCancel(const std::shared_ptr<AddGoalHandle> goal_handle);
  void handleAddExecute(const std::shared_ptr<AddGoalHandle> goal_handle);
  rclcpp_action::GoalResponse
  handleRemoveGoal(const rclcpp_action::GoalUUID &uuid,
                   std::shared_ptr<const RemoveCollisionObject::Goal> goal);
  rclcpp_action::CancelResponse
  handleRemoveCancel(const std::shared_ptr<RemoveGoalHandle> goal_handle);
  void handleRemoveExecute(const std::shared_ptr<RemoveGoalHandle> goal_handle);
  rclcpp_action::GoalResponse
  handleCheckExistGoal(const rclcpp_action::GoalUUID &uuid,
                       std::shared_ptr<const CheckObjectExists::Goal> goal);
  rclcpp_action::CancelResponse
  handleCheckExistCancel(const std::shared_ptr<CheckGoalHandle> goal_handle);
  void
  handleCheckExistExecute(const std::shared_ptr<CheckGoalHandle> goal_handle);

  rclcpp_action::GoalResponse
  handleAttachDetachGoal(const rclcpp_action::GoalUUID &uuid,
                         std::shared_ptr<const AttachDetachObject::Goal> goal);

  rclcpp_action::CancelResponse handleAttachDetachCancel(
      const std::shared_ptr<AttachDetachGoalHandle> goal_handle);

  void handleAttachDetachExecute(
      const std::shared_ptr<AttachDetachGoalHandle> goal_handle);

  rclcpp_action::GoalResponse
  handleGetObjectPoseGoal(const rclcpp_action::GoalUUID &uuid,
                          std::shared_ptr<const GetObjectPose::Goal> goal);

  rclcpp_action::CancelResponse handleGetObjectPoseCancel(
      const std::shared_ptr<GetObjectPoseGoalHandle> goal_handle);

  void handleGetObjectPoseExecute(
      const std::shared_ptr<GetObjectPoseGoalHandle> goal_handle);

  moveit_msgs::msg::CollisionObject createCollisionObject(
      const std::string &id, const std::string &shape,
      const std::vector<double> &dimensions, const std::string &frame_id,
      const geometry_msgs::msg::Pose &pose, const std::string &mesh_file,
      const double scale_mesh_x, const double scale_mesh_y,
      const double scale_mesh_z) const;

  bool attachedObjectExists(const std::string &id,
                            const std::string &link_name = "");

  std::optional<std::string>
  getAttachedObjectLinkById(const std::string &object_id);

  bool objectExists(const std::string &id);

  std::optional<moveit_msgs::msg::CollisionObject>
  getObjectDataById(const std::string &object_id);

  void printPlanningScene();

  bool
  validateShapeAndDimensions(const std::string &shape,
                             const std::vector<double> &dimensions,
                             const std::optional<std::string> &mesh_file) const;

  rclcpp::Publisher<moveit_msgs::msg::CollisionObject>::SharedPtr
      collision_object_publisher_;

  rclcpp::Publisher<moveit_msgs::msg::AttachedCollisionObject>::SharedPtr
      attached_collision_object_publisher_;
  rclcpp::Client<moveit_msgs::srv::GetPlanningScene>::SharedPtr
      get_planning_scene_client_;

  rclcpp_action::Server<AddCollisionObject>::SharedPtr
      add_object_action_server_;

  rclcpp_action::Server<RemoveCollisionObject>::SharedPtr
      remove_object_action_server_;

  rclcpp_action::Server<CheckObjectExists>::SharedPtr
      check_object_exists_action_server_;

  rclcpp_action::Server<AttachDetachObject>::SharedPtr
      attach_detach_object_action_server_;

  rclcpp_action::Server<GetObjectPose>::SharedPtr
      get_object_pose_action_server_;

  std::string frame_id_;

  rclcpp::CallbackGroup::SharedPtr service_callback_group_;
  rclcpp::CallbackGroup::SharedPtr action_callback_group_;
};

} // namespace rbs_mill_utils
