#include "rbs_mill_utils/collision_spawner.hpp"
#include <chrono>
#include <filesystem>
#include <memory>
#include <rclcpp/executors.hpp>
#include <rclcpp/logging.hpp>
#include <rclcpp/utilities.hpp>
#include <rclcpp_action/create_client.hpp>
#include <tf2/LinearMath/Quaternion.hpp>
#include <yaml-cpp/yaml.h>

CollisionSpawner::CollisionSpawner() : Node("collision_spawner") {
  this->declare_parameter("config_file", "");
  config_file_ = this->get_parameter("config_file").as_string();

  if (config_file_.empty()) {
    RCLCPP_FATAL(this->get_logger(), "Parameter 'config_file' if missing");
    rclcpp::shutdown();
    return;
  }

  RCLCPP_FATAL(this->get_logger(), "Starting loading yaml");

  try {
    _places_config =
        std::make_shared<PlacesConfig>(config_file_, this->get_logger());
  } catch (const std::exception &ex) {
    RCLCPP_FATAL(this->get_logger(), "Failure because: %s", ex.what());
  }

  RCLCPP_FATAL(this->get_logger(), "Ended loading yaml");

  if (_places_config->get()->models.empty()) {
    RCLCPP_WARN(this->get_logger(), "No valid objects to spawn");
    return;
  }

  add_object_action_client_ = rclcpp_action::create_client<AddCollisionObject>(
      this, "add_collision_object");
  remove_object_action_client_ =
      rclcpp_action::create_client<RemoveCollisionObject>(
          this, "remove_collision_object");
  check_object_exists_action_client_ =
      rclcpp_action::create_client<CheckObjectExists>(this,
                                                      "check_object_exists");

  if (!add_object_action_client_->wait_for_action_server(
          std::chrono::seconds(10))) {
    RCLCPP_FATAL(this->get_logger(),
                 "AddCollisionObject action server not available. shutdown");
    rclcpp::shutdown();
    return;
  }

  if (!remove_object_action_client_->wait_for_action_server(
          std::chrono::seconds(10))) {
    RCLCPP_FATAL(this->get_logger(),
                 "RemoveCollisionObject action server not available. shutdown");
    rclcpp::shutdown();
    return;
  }

  if (!check_object_exists_action_client_->wait_for_action_server(
          std::chrono::seconds(10))) {
    RCLCPP_WARN(this->get_logger(), "CheckObjectExist is not available");
  }

  removeExistingObjects();
  spawnObjects();
}

// bool CollisionSpawner::loadObjectsFromYaml(const std::string &filepath) {
//   YAML::Node places_config_node = YAML::LoadFile(filepath);
//
//   std::string model_dir;
//   if (auto node = places_config_node["model_dir"]; node && node.IsScalar()) {
//     objects_to_spawn_.model_dir = node.as<std::string>();
//   }
//
//   if (!places_config_node["models"] ||
//       !places_config_node["models"].IsSequence()) {
//     RCLCPP_FATAL(this->get_logger(),
//                  "No valid 'models' section found in config_file");
//     return false;
//   }
//   for (const auto &model : places_config_node["models"]) {
//     ObjectSpec object;
//     if (model["ps_ignore"] && model["ps_ignore"].as<bool>()) {
//       continue;
//     }
//     object.name = model["name"].as<std::string>();
//     object.type = model["type"].as<std::string>();
//     const auto &type = object.type;
//     if (type == "box" || type == "sphere" || type == "cylinder") {
//       if (!model["dimensions"]) {
//         RCLCPP_ERROR(this->get_logger(), "Model '%s' is missing
//         'dimensions'",
//                      object.name.c_str());
//         return false;
//       }
//       auto d = model["dimensions"].as<std::vector<double>>();
//
//       if (type == "box" && d.size() == 3) {
//         object.dimensions = d;
//       } else if (type == "sphere" && d.size() == 1) {
//         object.dimensions = d;
//       } else if (type == "cylinder" && d.size() == 2) {
//         object.dimensions = d;
//       } else {
//         RCLCPP_ERROR(this->get_logger(),
//                      "Invalid dimensions count for type '%s'", type.c_str());
//         return false;
//       }
//
//     } else if (type == "mesh") {
//       if (!model["mesh_filename"]) {
//         RCLCPP_ERROR(this->get_logger(), "mesh object missing
//         mesh_filename"); return false;
//       }
//       std::string mesh_filename = model["mesh_filename"].as<std::string>();
//       if (objects_to_spawn_.model_dir.has_value()) {
//         object.mesh_filepath =
//             std::filesystem::path(objects_to_spawn_.model_dir->c_str()) /
//             object.name / "meshes" / mesh_filename;
//         RCLCPP_WARN(this->get_logger(), "mesh filepath is: %s",
//                     object.mesh_filepath->c_str());
//       } else {
//         RCLCPP_ERROR(this->get_logger(),
//                      "mesh_filename in empty in config_file");
//         return false;
//       }
//
//     } else {
//       RCLCPP_ERROR(this->get_logger(), "Unrecognized object type: '%s'",
//                    type.c_str());
//       return false;
//     }
//     if (model["frame_id"]) {
//       object.frame_id = model["frame_id"].as<std::string>();
//     }
//     if (model["pose"]) {
//       auto pose = model["pose"];
//
//       if (pose["position"]) {
//         auto position = model["pose"]["position"];
//         object.pose.position.x = position["x"].as<double>();
//         object.pose.position.y = position["y"].as<double>();
//         object.pose.position.z = position["z"].as<double>();
//       }
//       if (pose["orientation"]) {
//         auto orientation = model["pose"]["orientation"];
//         if (!isQuat(orientation)) {
//           auto quat = rpyToQuat(orientation["r"].as<double>(),
//                                 orientation["p"].as<double>(),
//                                 orientation["y"].as<double>());
//
//           object.pose.orientation.x = quat.x;
//           object.pose.orientation.y = quat.y;
//           object.pose.orientation.z = quat.z;
//           object.pose.orientation.w = quat.w;
//         } else {
//           object.pose.orientation.x = orientation["x"].as<double>();
//           object.pose.orientation.y = orientation["y"].as<double>();
//           object.pose.orientation.z = orientation["z"].as<double>();
//           object.pose.orientation.w = orientation["w"].as<double>();
//         }
//       }
//     }
//
//     objects_to_spawn_.models.push_back(object);
//   }
//
//   return true;
// }

void CollisionSpawner::removeExistingObjects() {
  RCLCPP_INFO(this->get_logger(), "Removing existing objects...");
  for (const auto &obj : _places_config->get()->models) {
    auto object_id = obj.name;

    // 1. Check if it exists
    if (!isObjectExist(object_id)) {
      RCLCPP_INFO(this->get_logger(),
                  "Object '%s' does not exist. Skipping removal.",
                  object_id.c_str());
      continue;
    }

    // 2. Proceed with remove
    sendRemoveGoal(object_id);
  }
}

void CollisionSpawner::sendRemoveGoal(const std::string &object_id) {
  auto goal = RemoveCollisionObject::Goal();
  goal.id = object_id;

  RCLCPP_INFO(this->get_logger(), "Sending remove goal for object '%s'.",
              object_id.c_str());

  auto send_goal_future = remove_object_action_client_->async_send_goal(goal);

  // Wait for the goal to be accepted
  if (rclcpp::spin_until_future_complete(
          this->get_node_base_interface(), send_goal_future,
          std::chrono::seconds(5)) != rclcpp::FutureReturnCode::SUCCESS) {
    RCLCPP_WARN(this->get_logger(),
                "Timeout while sending remove goal for object '%s'.",
                object_id.c_str());
    return;
  }

  auto goal_handle = send_goal_future.get();
  if (!goal_handle) {
    RCLCPP_ERROR(this->get_logger(),
                 "Remove goal was rejected for object '%s'.",
                 object_id.c_str());
    return;
  }

  // Wait for the result
  auto get_result_future =
      remove_object_action_client_->async_get_result(goal_handle);

  if (rclcpp::spin_until_future_complete(
          this->get_node_base_interface(), get_result_future,
          std::chrono::seconds(5)) != rclcpp::FutureReturnCode::SUCCESS) {
    RCLCPP_WARN(this->get_logger(),
                "Timeout while waiting for remove result for object '%s'.",
                object_id.c_str());
    return;
  }

  auto result = get_result_future.get();
  switch (result.code) {
  case rclcpp_action::ResultCode::SUCCEEDED:
    RCLCPP_INFO(this->get_logger(), "Successfully removed object '%s'.",
                object_id.c_str());
    break;
  case rclcpp_action::ResultCode::ABORTED:
    RCLCPP_WARN(this->get_logger(),
                "Remove action was aborted for object '%s'.",
                object_id.c_str());
    break;
  case rclcpp_action::ResultCode::CANCELED:
    RCLCPP_WARN(this->get_logger(),
                "Remove action was canceled for object '%s'.",
                object_id.c_str());
    break;
  default:
    RCLCPP_WARN(this->get_logger(),
                "Unknown result code for remove action of object '%s'.",
                object_id.c_str());
    break;
  }
}

void CollisionSpawner::spawnObjects() {
  for (const auto &obj : _places_config->get()->models) {
    sendAddGoal(obj);
  }
}

void CollisionSpawner::sendAddGoal(const PlacesConfig::PlaceSpec &obj_spec) {

  auto goal = AddCollisionObject::Goal();
  goal.id = obj_spec.name;
  goal.shape = obj_spec.type;
  goal.dimensions =
      obj_spec.type == "mesh" ? std::vector<double>() : obj_spec.dimensions;
  goal.pose = obj_spec.pose;

  if (obj_spec.type == "mesh") {
    if (!obj_spec.mesh_filepath.has_value()) {
      RCLCPP_FATAL(this->get_logger(),
                   "Mesh filepath has no value for name: %s",
                   obj_spec.name.c_str());
    }
    goal.mesh_file = obj_spec.mesh_filepath.value();
    auto [sx, sy, sz] = obj_spec.scale_xyz;
    goal.scale_mesh_x = sx;
    goal.scale_mesh_y = sy;
    goal.scale_mesh_z = sz;
  }

  RCLCPP_INFO(this->get_logger(), "Sending add goal for object '%s'.",
              obj_spec.name.c_str());

  auto send_goal_future = add_object_action_client_->async_send_goal(goal);

  // Wait for the goal to be accepted
  if (rclcpp::spin_until_future_complete(
          this->get_node_base_interface(), send_goal_future,
          std::chrono::seconds(5)) != rclcpp::FutureReturnCode::SUCCESS) {
    RCLCPP_WARN(this->get_logger(),
                "Timeout while sending add goal for object '%s'.",
                obj_spec.name.c_str());
    return;
  }

  auto goal_handle = send_goal_future.get();
  if (!goal_handle) {
    RCLCPP_ERROR(this->get_logger(), "Add goal was rejected for object '%s'.",
                 obj_spec.name.c_str());
    return;
  }

  // Wait for the result
  auto get_result_future =
      add_object_action_client_->async_get_result(goal_handle);

  if (rclcpp::spin_until_future_complete(
          this->get_node_base_interface(), get_result_future,
          std::chrono::seconds(5)) != rclcpp::FutureReturnCode::SUCCESS) {
    RCLCPP_WARN(this->get_logger(),
                "Timeout while waiting for add result for object '%s'.",
                obj_spec.name.c_str());
    return;
  }

  auto result = get_result_future.get();
  switch (result.code) {
  case rclcpp_action::ResultCode::SUCCEEDED:
    RCLCPP_INFO(this->get_logger(), "Successfully added object '%s'.",
                obj_spec.name.c_str());
    break;
  case rclcpp_action::ResultCode::ABORTED:
    RCLCPP_WARN(this->get_logger(), "Add action was aborted for object '%s'.",
                obj_spec.name.c_str());
    break;
  case rclcpp_action::ResultCode::CANCELED:
    RCLCPP_WARN(this->get_logger(), "Add action was canceled for object '%s'.",
                obj_spec.name.c_str());
    break;
  default:
    RCLCPP_WARN(this->get_logger(),
                "Unknown result code for add action of object '%s'.",
                obj_spec.name.c_str());
    break;
  }
}

bool CollisionSpawner::isObjectExist(const std::string &object_id) {
  if (!check_object_exists_action_client_->action_server_is_ready()) {
    RCLCPP_WARN(this->get_logger(),
                "CheckObjectExists action server not ready. Assuming object "
                "'%s' does not exist.",
                object_id.c_str());
    return false;
  }

  auto goal_msg = CheckObjectExists::Goal();
  goal_msg.object_id = object_id;

  RCLCPP_INFO(this->get_logger(),
              "Sending CheckObjectExists goal for object '%s'.",
              object_id.c_str());

  auto send_goal_future =
      check_object_exists_action_client_->async_send_goal(goal_msg);

  // Wait for the goal to be accepted
  if (rclcpp::spin_until_future_complete(
          this->get_node_base_interface(), send_goal_future,
          std::chrono::seconds(3)) != rclcpp::FutureReturnCode::SUCCESS) {
    RCLCPP_WARN(this->get_logger(),
                "Timeout while sending CheckObjectExists goal for object '%s'. "
                "Assuming it does not exist.",
                object_id.c_str());
    return false;
  }

  auto goal_handle = send_goal_future.get();
  if (!goal_handle) {
    RCLCPP_ERROR(this->get_logger(),
                 "CheckObjectExists goal was rejected for object '%s'. "
                 "Assuming it does not exist.",
                 object_id.c_str());
    return false;
  }

  // Wait for the result
  auto get_result_future =
      check_object_exists_action_client_->async_get_result(goal_handle);

  if (rclcpp::spin_until_future_complete(
          this->get_node_base_interface(), get_result_future,
          std::chrono::seconds(3)) != rclcpp::FutureReturnCode::SUCCESS) {
    RCLCPP_WARN(this->get_logger(),
                "Timeout while waiting for CheckObjectExists result for object "
                "'%s'. Assuming it does not exist.",
                object_id.c_str());
    return false;
  }

  auto result = get_result_future.get().result;
  RCLCPP_INFO(this->get_logger(),
              "CheckObjectExists result for object '%s': exists=%s",
              object_id.c_str(), result->ok ? "true" : "false");
  return result->ok;
}

// bool CollisionSpawner::isQuat(const YAML::Node &node) {
//   if (node["x"] && node["y"] && node["z"] && node["w"]) {
//     return true;
//   } else if (node["r"] && node["p"] && node["y"]) {
//     return false;
//   }
//   throw std::invalid_argument("Provided YAML Node is not quat or euler
//   angles");
// }

// geometry_msgs::msg::Quaternion
// CollisionSpawner::rpyToQuat(double roll, double pitch, double yaw) {
//   tf2::Quaternion q;
//   q.setRPY(roll, pitch, yaw);
//   q.normalize();
//   geometry_msgs::msg::Quaternion q_msg;
//   q_msg.x = q.x();
//   q_msg.y = q.y();
//   q_msg.z = q.z();
//   q_msg.w = q.w();
//   return q_msg;
// }

int main(int argc, char *argv[]) {
  rclcpp::init(argc, argv);
  auto node = std::make_shared<CollisionSpawner>();
  rclcpp::spin(node);
  rclcpp::shutdown();
  return 0;
}
